"""
Module for handling Lennard-Jones interaction
"""
import numpy as np
import numba as nb


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential : sarkas.potentials.core.Potential
        Class handling potential form.

    params : sarkas.core.Parameters
        Simulation's parameters.

    """
    potential.matrix = np.zeros((4, params.num_species, params.num_species))
    # See Lima Physica A 391 4281 (2012) for the following definitions
    if not hasattr(potential, 'powers'):
        potential.powers = np.array([12, 6])

    exponent = potential.powers[0] / (potential.powers[1] - potential.powers[0])
    lj_constant = potential.powers[1]/(potential.powers[0] - potential.powers[1])
    lj_constant *= (potential.powers[1]/potential.powers[0]) ** exponent

    # Use the Lorentz-Berthelot mixing rules.
    # Lorentz: sigma_12 = 0.5 * (sigma_1 + sigma_2)
    # Berthelot: epsilon_12 = sqrt( eps_1 eps2)
    sigma2 = 0.0
    epsilon_tot = 0.0
    # Recall that species_charges contains sqrt(epsilon)
    for i, q1 in enumerate(params.species_charges):
        for j, q2 in enumerate(params.species_charges):
            potential.matrix[0, i, j] = lj_constant * q1 * q2
            potential.matrix[1, i, j] = 0.5 * (params.species_lj_sigmas[i] + params.species_lj_sigmas[j])
            potential.matrix[2, i, j] = potential.powers[0]
            potential.matrix[3, i, j] = potential.powers[1]
            sigma2 += params.species_lj_sigmas[i]
            epsilon_tot += q1 * q2

    potential.force = lj_force

    # Calculate the force error
    deriv_exp = (potential.powers[0] + 1)
    params.force_error = np.sqrt(np.pi * sigma2 ** potential.powers[0])
    params.force_error /= np.sqrt(deriv_exp * potential.rc ** deriv_exp)
    params.force_error *= np.sqrt(params.total_num_ptcls / params.box_volume) * params.a_ws ** 2


@nb.njit
def lj_force(r, pot_matrix_ij):
    """
    Calculates the PP force between particles using Lennard-Jones Potential.
    
    Parameters
    ----------
    pot_matrix_ij : array
        LJ potential parameters. 

    r : float
        Particles' distance.


    Returns
    -------
    U : float
        Potential.

    force : float
        Force.
    """
    """
    Notes
    -----
    pot_matrix[0] = epsilon_12 * lj_constant
    pot_matrix[1] = sigmas
    pot_matrix[2] = highest power
    pot_matrix[3] = lowest power
    
    """

    epsilon = pot_matrix_ij[0]
    sigma = pot_matrix_ij[1]
    s_over_r = sigma / r
    s_over_r_high = s_over_r ** pot_matrix_ij[2]
    s_over_r_low = s_over_r ** pot_matrix_ij[3]

    U = epsilon * (s_over_r_high - s_over_r_low)
    force = epsilon * (pot_matrix_ij[2] * s_over_r_high - pot_matrix_ij[3] * s_over_r_low) / r

    return U, force



""" 
Module for handling Coulomb interaction
"""

import numpy as np
from numba import njit
import math as mt


def update_params(potential, params):
    """
    Assign potential dependent simulation's parameters.

    Parameters
    ----------
    potential: sarkas.potentials.core.Potential
        Potential's information

    params: sarkas.core.Parameters
        Simulation's parameters

    """
    """
    Dev Notes:
    -----
    Coulomb_matrix[0,i,j] : qi qj/(4pi esp0) Force factor between two particles.
    Coulomb_matrix[1,i,j] : Ewald parameter in the case of P3M Algorithm. Same value for all species
    """
    potential.matrix = np.zeros((2, params.num_species, params.num_species))

    for i, q1 in enumerate(params.species_charges):
        for j, q2 in enumerate(params.species_charges):
            potential.matrix[0, i, j] = q1 * q2 / params.fourpie0

    potential.matrix[1, :, :] = potential.pppm_alpha_ewald
    # Calculate the (total) plasma frequency
    potential.force = coulomb_force_pppm

    # PP force error calculation. Note that the equation was derived for a single component plasma.
    alpha_times_rcut = - (potential.pppm_alpha_ewald * potential.rc) ** 2
    params.pppm_pp_err = 2.0 * np.exp(alpha_times_rcut) / np.sqrt(potential.rc)
    params.pppm_pp_err *= np.sqrt(params.total_num_ptcls) * params.a_ws ** 2 / np.sqrt(params.box_volume)


@njit
def coulomb_force_pppm(r, pot_matrix):
    """ 
    Calculate Potential and Force between two particles when the P3M algorithm is chosen.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : numpy.ndarray
        It contains potential dependent variables.

    Returns
    -------
    U_s_r : float
        Potential value.
                
    fr : float
        Force between two particles. 
    
    """

    alpha = pot_matrix[1]  # Ewald parameter alpha
    alpha_r = alpha * r
    r2 = r * r
    U = pot_matrix[0] * mt.erfc(alpha_r) / r
    f1 = mt.erfc(alpha_r) / r2
    f2 = (2.0 * alpha / np.sqrt(np.pi) / r) * np.exp(- alpha_r ** 2)
    fr = pot_matrix[0] * (f1 + f2)

    return U, fr


@njit
def coulomb_force(r, pot_matrix):
    """
    Calculate the coulomb potential and force between two particles.

    Parameters
    ----------
    r : float
        Distance between two particles.

    pot_matrix : numpy.ndarray
        It contains potential dependent variables.

    Returns
    -------
    U : float
        Potential value.

    fr : float
        Force between two particles.

    """

    U = pot_matrix[0] / r
    fr = U / r

    return U, fr



"""
Module of various types of time_evolution
"""

import numpy as np
from numba import njit
from IPython import get_ipython

if get_ipython().__class__.__name__ == 'ZMQInteractiveShell':
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm
# import fmm3dpy as fmm
# from sarkas.potentials import force_pm, force_pp


class Integrator:
    """
    Class used to assign integrator type.

    Attributes
    ----------
    dt: float
        Timestep.

    equilibration_steps: int
        Total number of equilibration timesteps.

    eq_dump_step: int
        Equilibration dump interval.

    kB: float
        Boltzmann constant.

    magnetized: bool
        Magnetized simulation flag.

    production_steps: int
        Total number of production timesteps.

    prod_dump_step: int
        Production dump interval.

    species_num: numpy.ndarray
        Number of particles of each species. copy of ``parameters.species_num``.

    box_lengths: numpy.ndarray
        Length of each box side.

    verbose: bool
        Verbose output flag.

    type: str
        Integrator type.

    update: func
        Integrator choice. 'verlet' or 'magnetic_verlet'.

    update_accelerations: func
        Link to the correct potential update function.

    thermostate: func
        Link to the correct thermostat function.

    """

    def __init__(self):
        self.type = None
        self.dt = None
        self.kB = None
        self.magnetized = False
        self.electrostatic_equilibration = False
        self.production_steps = None
        self.equilibration_steps = None
        self.magnetization_steps = None
        self.prod_dump_step = None
        self.eq_dump_step = None
        self.mag_dump_steps = None
        self.update = None
        self.species_num = None
        self.box_lengths = None
        self.verbose = False
        self.supported_integrators = ['verlet', 'verlet_langevin', 'magnetic_verlet', 'magnetic_boris']

    # def __repr__(self):
    #     sortedDict = dict(sorted(self.__dict__.items(), key=lambda x: x[0].lower()))
    #     disp = 'Integrator( \n'
    #     for key, value in sortedDict.items():
    #         disp += "\t{} : {}\n".format(key, value)
    #     disp += ')'
    #     return disp

    def from_dict(self, input_dict: dict):
        """
        Update attributes from input dictionary.

        Parameters
        ----------
        input_dict: dict
            Dictionary to be copied.

        """
        self.__dict__.update(input_dict)

    def setup(self, params, thermostat, potential):
        """
        Assign attributes from simulation's parameters and classes.

        Parameters
        ----------
        params: sarkas.core.parameters
            Parameters class.

        thermostat: sarkas.time_evolution.thermostat
            Thermostat class

        potential: sarkas.potentials.core.Potential
            Potential class.

        """
        self.box_lengths = np.copy(params.box_lengths)
        self.kB = params.kB
        self.species_num = np.copy(params.species_num)
        self.verbose = params.verbose

        if self.dt is None:
            self.dt = params.dt

        if self.production_steps is None:
            self.production_steps = params.production_steps

        if self.equilibration_steps is None:
            self.equilibration_steps = params.equilibration_steps

        if self.prod_dump_step is None:
            if hasattr(params, 'prod_dump_step'):
                self.prod_dump_step = params.prod_dump_step
            else:
                self.prod_dump_step = int(0.1 * self.production_steps)

        if self.eq_dump_step is None:
            if hasattr(params, 'eq_dump_step'):
                self.eq_dump_step = params.eq_dump_step
            else:
                self.eq_dump_step = int(0.1 * self.equilibration_steps)

        assert self.type.lower() in self.supported_integrators, 'Wrong integrator choice.'

        # Assign integrator.update to the correct method
        if self.type.lower() == "verlet":
            self.update = self.verlet

        elif self.type.lower() == "verlet_langevin":

            self.sigma = np.sqrt(
                2. * self.langevin_gamma * params.kB * params.species_temperatures / params.species_masses)
            self.c1 = (1. - 0.5 * self.langevin_gamma * self.dt)
            self.c2 = 1. / (1. + 0.5 * self.langevin_gamma * self.dt)
            self.update = self.verlet_langevin

        elif self.type.lower() == "magnetic_verlet":

            # Create the unit vector of the magnetic field
            self.magnetic_field_uvector = params.magnetic_field / np.linalg.norm(params.magnetic_field)
            self.omega_c = np.zeros((params.total_num_ptcls, params.dimensions))

            sp_start = 0
            sp_end = 0
            for ic, sp_np in enumerate(params.species_num):
                sp_end += sp_np
                self.omega_c[sp_start: sp_end, :] = params.species_cyclotron_frequencies[ic]
                sp_start += sp_np

            # Calculate functions for magnetic integrator
            # This could be used when the generalization to Forest-Ruth and MacLachlan algorithms will be implemented
            # In a magnetic Velocity-Verlet the coefficient is 1/2, see eq.~(78) in :cite:`Chin2008`
            self.magnetic_helpers(0.5)

            # array to temporary store velocities
            # Luciano: I have the vague doubt that allocating memory for these arrays is faster than calculating them
            # each time step
            self.v_B = np.zeros((params.total_num_ptcls, params.dimensions))
            self.v_F = np.zeros((params.total_num_ptcls, params.dimensions))

            if np.dot(self.magnetic_field_uvector, np.array([0.0, 0.0, 1.0])) == 1.0:
                self.update = self.magnetic_verlet_zdir
            else:
                self.update = self.magnetic_verlet

        elif self.type.lower() == "magnetic_boris":

            # Create the unit vector of the magnetic field
            self.magnetic_field_uvector = params.magnetic_field / np.linalg.norm(params.magnetic_field)
            self.omega_c = np.zeros((params.total_num_ptcls, params.dimensions))

            sp_start = 0
            sp_end = 0
            for ic, sp_np in enumerate(params.species_num):
                sp_end += sp_np
                self.omega_c[sp_start: sp_end, :] = params.species_cyclotron_frequencies[ic]
                sp_start += sp_np

            # In a leapfrog-type algorithm the coefficient is different for the acceleration and magnetic rotation
            # see eq.~(79) in :cite:`Chin2008`
            # self.magnetic_helpers(1.0)

            if np.dot(self.magnetic_field_uvector, np.array([0.0, 0.0, 1.0])) == 1.0:
                self.update = self.magnetic_boris_zdir
            else:
                self.update = self.magnetic_boris

            # array to temporary store velocities
            # Luciano: I have the vague doubt that allocating memory for these arrays is faster than calculating them
            # each time step
            self.v_B = np.zeros((params.total_num_ptcls, params.dimensions))
            self.v_F = np.zeros((params.total_num_ptcls, params.dimensions))

        if params.magnetized:
            self.magnetized = True

            if self.electrostatic_equilibration:
                params.electrostatic_equilibration = True
                self.magnetic_integrator = self.update
                self.update = self.verlet

                if self.magnetization_steps is None:
                    self.magnetization_steps = params.magnetization_steps

                if self.prod_dump_step is None:
                    if hasattr(params, 'mag_dump_step'):
                        self.mag_dump_step = params.mag_dump_step
                    else:
                        self.mag_dump_step = int(0.1 * self.production_steps)

        if not potential.method == 'FMM':
            if potential.pppm_on:
                self.update_accelerations = potential.update_pppm
            else:
                if potential.linked_list_on:
                    self.update_accelerations = potential.update_linked_list
                else:
                    self.update_accelerations = potential.update_brute
        else:
            self.update_accelerations = potential.update_fmm

        self.thermostate = thermostat.update

    def equilibrate(self, it_start, ptcls, checkpoint):
        """
        Loop over the equilibration steps.

        Parameters
        ----------
        it_start: int
            Initial step of equilibration.

        ptcls: sarkas.core.Particles
            Particles' class.

        checkpoint: sarkas.utilities.InputOutput
            IO class for saving dumps.

        """

        for it in tqdm(range(it_start, self.equilibration_steps), disable=not self.verbose):
            # Calculate the Potential energy and update particles' data
            self.update(ptcls)
            if (it + 1) % self.eq_dump_step == 0:
                checkpoint.dump('equilibration', ptcls, it + 1)
            self.thermostate(ptcls, it)
        ptcls.remove_drift()

    def magnetize(self, it_start, ptcls, checkpoint):
        self.update = self.magnetic_integrator
        for it in tqdm(range(it_start, self.magnetization_steps), disable=not self.verbose):
            # Calculate the Potential energy and update particles' data
            self.update(ptcls)
            if (it + 1) % self.mag_dump_step == 0:
                checkpoint.dump('magnetization', ptcls, it + 1)
            self.thermostate(ptcls, it)

    def produce(self, it_start, ptcls, checkpoint):
        """
        Loop over the production steps.

        Parameters
        ----------
        it_start: int
            Initial step of production phase.

        ptcls: sarkas.core.Particles
            Particles' class.

        checkpoint: sarkas.utilities.InputOutput
            IO class for saving dumps.

        """
        for it in tqdm(range(it_start, self.production_steps), disable=(not self.verbose)):

            # Move the particles and calculate the potential
            self.update(ptcls)
            if (it + 1) % self.prod_dump_step == 0:
                # Save particles' data for restart
                checkpoint.dump('production', ptcls, it + 1)

    def verlet_langevin(self, ptcls):
        """
        Update particles class using the velocity verlet algorithm and Langevin damping.

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.


        """
        beta = ptcls.gaussian(0., 1., ptcls.pos.shape[0])
        sp_start = 0  # start index for species loop
        sp_end = 0
        for ic, num in enumerate(self.species_num):
            sp_end += num
            ptcls.pos[sp_start:sp_end, :] += self.c1 * self.dt * ptcls.vel[sp_start:sp_end, :] \
                                             + 0.5 * self.dt ** 2 * ptcls.acc[sp_start:sp_end, :] \
                                             + 0.5 * self.sigma[ic] * self.dt ** 1.5 * beta

        # Periodic boundary condition
        enforce_pbc(ptcls.pos, ptcls.pbc_cntr, self.box_lengths)

        acc_old = np.copy(ptcls.acc)
        self.update_accelerations(ptcls)

        sp_start = 0
        sp_end = 0
        for ic, num in enumerate(self.species_num):
            sp_end += num

            ptcls.vel[sp_start:sp_end, :] = self.c1 * self.c2 * ptcls.vel[sp_start:sp_end, :] \
                                            + 0.5 * self.c2 * self.dt * (ptcls.acc[sp_start:sp_end, :]
                                                                         + acc_old[sp_start:sp_end, :]) \
                                            + self.c2 * self.sigma[ic] * np.sqrt(self.dt) * beta
            sp_start = sp_end

    def verlet(self, ptcls):
        """
        Update particles' class based on velocity verlet algorithm.
        More information can be found here: https://en.wikipedia.org/wiki/Verlet_integration
        or on the Sarkas website.

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.

        """
        # First half step velocity update
        ptcls.vel += 0.5 * ptcls.acc * self.dt
        # Full step position update
        ptcls.pos += ptcls.vel * self.dt

        # Periodic boundary condition
        enforce_pbc(ptcls.pos, ptcls.pbc_cntr, self.box_lengths)
        # Compute total potential energy and acceleration for second half step velocity update
        self.update_accelerations(ptcls)

        # Second half step velocity update
        ptcls.vel += 0.5 * ptcls.acc * self.dt

    def magnetic_helpers(self, coefficient):
        """Calculate the trigonometric functions of the magnetic integrators.

        Parameters
        ----------
        coefficient: float
            Timestep coefficient.

        Notes
        -----
        This is useful for the Leapfrog magnetic algorithm and future Forest-Ruth and MacLachlan algorithms.

        """
        theta = self.omega_c * self.dt * coefficient
        self.sdt = np.sin(theta)
        self.cdt = np.cos(theta)
        self.ccodt = 1.0 - self.cdt
        self.ssodt = 1.0 - self.sdt / theta

    def magnetic_verlet_zdir(self, ptcls):
        """
        Update particles' class based on velocity verlet method in the case of a
        constant magnetic field along the :math:`z` axis. For more info see eq. (78) of Ref. :cite:`Chin2008`

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.

        Returns
        -------
        potential_energy : float
             Total potential energy.

        Notes
        -----
        This integrator is faster than `magnetic_verlet` but valid only for a magnetic field in the :math:`z`-direction.
        This is the preferred choice in this case.
        """

        # First half step of velocity update
        # # Magnetic rotation x - velocity
        # (B x v)_x  = -v_y, (B x B x v)_x = -v_x
        self.v_B[:, 0] = ptcls.vel[:, 1] * self.sdt[:, 0] + ptcls.vel[:, 0] * self.cdt[:, 0]
        # Magnetic rotation y - velocity
        # (B x v)_y  = v_x, (B x B x v)_y = -v_y
        self.v_B[:, 1] = - ptcls.vel[:, 0] * self.sdt[:, 0] + ptcls.vel[:, 1] * self.cdt[:, 1]

        # Magnetic + Const force field x - velocity
        # (B x a)_x  = -a_y, (B x B x a)_x = -a_x
        self.v_F[:, 0] = self.ccodt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1] \
                         + self.sdt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0]
        # Magnetic + Const force field y - velocity
        # (B x a)_y  = a_x, (B x B x a)_y = -a_y
        self.v_F[:, 1] = - self.ccodt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0] \
                         + self.sdt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1]

        ptcls.vel[:, 0] = self.v_B[:, 0] + self.v_F[:, 0]
        ptcls.vel[:, 1] = self.v_B[:, 1] + self.v_F[:, 1]
        ptcls.vel[:, 2] += 0.5 * self.dt * ptcls.acc[:, 2]

        # Position update
        ptcls.pos += ptcls.vel * self.dt

        # Periodic boundary condition
        enforce_pbc(ptcls.pos, ptcls.pbc_cntr, self.box_lengths)

        # Compute total potential energy and acceleration for second half step velocity update
        potential_energy = self.update_accelerations(ptcls)

        # # Magnetic rotation x - velocity
        # (B x v)_x  = -v_y, (B x B x v)_x = -v_x
        self.v_B[:, 0] = ptcls.vel[:, 1] * self.sdt[:, 0] + ptcls.vel[:, 0] * self.cdt[:, 0]
        # Magnetic rotation y - velocity
        # (B x v)_y  = v_x, (B x B x v)_y = -v_y
        self.v_B[:, 1] = - ptcls.vel[:, 0] * self.sdt[:, 0] + ptcls.vel[:, 1] * self.cdt[:, 1]

        # Magnetic + Const force field x - velocity
        # (B x a)_x  = -a_y, (B x B x a)_x = -a_x
        self.v_F[:, 0] = self.ccodt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1] \
                         + self.sdt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0]
        # Magnetic + Const force field y - velocity
        # (B x a)_y  = a_x, (B x B x a)_y = -a_y
        self.v_F[:, 1] = - self.ccodt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0] \
                         + self.sdt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1]

        ptcls.vel[:, 0] = self.v_B[:, 0] + self.v_F[:, 0]
        ptcls.vel[:, 1] = self.v_B[:, 1] + self.v_F[:, 1]
        ptcls.vel[:, 2] += 0.5 * self.dt * ptcls.acc[:, 2]

        return potential_energy

    def magnetic_verlet(self, ptcls):
        """
        Update particles' class based on velocity verlet method in the case of an arbitrary direction of the
        constant magnetic field. For more info see eq. (78) of Ref. :cite:`Chin2008`

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.

        Returns
        -------
        potential_energy : float
             Total potential energy.

        Notes
        -----
        :cite:`Chin2008` equations are written for a negative charge. This allows him to write
        :math:`\dot{\mathbf v} = \omega_c \hat{B} \\times \mathbf v`. In the case of positive charges we will have
        :math:`\dot{\mathbf v} = - \omega_c \hat{B} \\times \mathbf v`. Hence the reason of the different signs in the
        formulas below compared to Chin's.

        Warnings
        --------
        This integrator is valid for a magnetic field in an arbitrary direction. However, while the integrator works for
        an arbitrary direction, methods in `sarkas.tool.observables` work only for a magnetic field in the
        :math:`z` - direction. Hence, if you choose to use this integrator remember to change your physical observables.

        """
        # Calculate the cross products
        b_cross_v = np.cross(self.magnetic_field_uvector, ptcls.vel)
        b_cross_b_cross_v = np.cross(self.magnetic_field_uvector, b_cross_v)
        b_cross_a = np.cross(self.magnetic_field_uvector, ptcls.acc)
        b_cross_b_cross_a = np.cross(self.magnetic_field_uvector, b_cross_a)

        # First half step of velocity update
        ptcls.vel += - self.sdt * b_cross_v + self.ccodt * b_cross_b_cross_v

        ptcls.vel += 0.5 * ptcls.acc * self.dt - self.ccodt / self.omega_c * b_cross_a \
                   + 0.5 * self.dt * self.ssodt * b_cross_b_cross_a

        # Position update
        ptcls.pos += ptcls.vel * self.dt

        # Periodic boundary condition
        enforce_pbc(ptcls.pos, ptcls.pbc_cntr, self.box_lengths)

        # Compute total potential energy and acceleration for second half step velocity update
        potential_energy = self.update_accelerations(ptcls)

        # Re-calculate the cross products
        b_cross_v = np.cross(self.magnetic_field_uvector, ptcls.vel)
        b_cross_b_cross_v = np.cross(self.magnetic_field_uvector, b_cross_v)
        b_cross_a = np.cross(self.magnetic_field_uvector, ptcls.acc)
        b_cross_b_cross_a = np.cross(self.magnetic_field_uvector, b_cross_a)

        # Second half step velocity update
        ptcls.vel += - self.sdt * b_cross_v + self.ccodt * b_cross_b_cross_v

        ptcls.vel += 0.5 * ptcls.acc * self.dt - self.ccodt / self.omega_c * b_cross_a \
                   + 0.5 * self.dt * self.ssodt * b_cross_b_cross_a

        return potential_energy

    def magnetic_boris_zdir(self, ptcls):
        """
        Update particles' class using the Boris algorithm in the case of a
        constant magnetic field along the :math:`z` axis. For more info see eqs. (80) - (81) of Ref. :cite:`Chin2008`

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.

        Returns
        -------
        potential_energy : float
             Total potential energy.

        """
        # First half step of velocity update: Apply exp(eV_BF)
        # epsilon/2 V_F
        self.magnetic_helpers(0.5)
        # Magnetic + Const force field x - velocity
        # (B x a)_x  = -a_y, (B x B x a)_x = -a_x
        self.v_F[:, 0] = self.ccodt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1] \
                         + self.sdt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0]
        # Magnetic + Const force field y - velocity
        # (B x a)_y  = a_x, (B x B x a)_y = -a_y
        self.v_F[:, 1] = - self.ccodt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0] \
                         + self.sdt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1]

        ptcls.vel[:, 0] += self.v_F[:, 0]
        ptcls.vel[:, 1] += self.v_F[:, 1]
        ptcls.vel[:, 2] += 0.5 * self.dt * ptcls.acc[:, 2]

        # epsilon V_B
        self.magnetic_helpers(1.0)
        # Magnetic rotation x - velocity
        # (B x v)_x  = -v_y, (B x B x v)_x = -v_x
        self.v_B[:, 0] = ptcls.vel[:, 1] * self.sdt[:, 0] + ptcls.vel[:, 0] * self.cdt[:, 0]
        # Magnetic rotation y - velocity
        # (B x v)_y  = v_x, (B x B x v)_y = -v_y
        self.v_B[:, 1] = - ptcls.vel[:, 0] * self.sdt[:, 0] + ptcls.vel[:, 1] * self.cdt[:, 1]

        ptcls.vel[:, 0] = np.copy(self.v_B[:, 0])
        ptcls.vel[:, 1] = np.copy(self.v_B[:, 1])

        # # epsilon/2 V_F
        self.magnetic_helpers(0.5)
        # Magnetic + Const force field x - velocity
        # (B x a)_x  = -a_y, (B x B x a)_x = -a_x
        self.v_F[:, 0] = self.ccodt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1] \
                         + self.sdt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0]
        # Magnetic + Const force field y - velocity
        # (B x a)_y  = a_x, (B x B x a)_y = -a_y
        self.v_F[:, 1] = - self.ccodt[:, 0] / self.omega_c[:, 0] * ptcls.acc[:, 0] \
                         + self.sdt[:, 1] / self.omega_c[:, 1] * ptcls.acc[:, 1]

        ptcls.vel[:, 0] += self.v_F[:, 0]
        ptcls.vel[:, 1] += self.v_F[:, 1]
        ptcls.vel[:, 2] += 0.5 * self.dt * ptcls.acc[:, 2]

        # Full step position update
        ptcls.pos += ptcls.vel * self.dt

        # Periodic boundary condition
        enforce_pbc(ptcls.pos, ptcls.pbc_cntr, self.box_lengths)

        # Compute total potential energy and acceleration for second half step velocity update
        potential_energy = self.update_accelerations(ptcls)

        return potential_energy

    def magnetic_boris(self, ptcls):
        """
        Update particles' class using the Boris algorithm in the case of a
        constant magnetic field along the :math:`z` axis. For more info see eqs. (80) - (81) of Ref. :cite:`Chin2008`

        Parameters
        ----------
        ptcls: sarkas.core.Particles
            Particles data.

        Returns
        -------
        potential_energy : float
             Total potential energy.

        """

        # First half step of velocity update: Apply exp(eV_BF)
        # epsilon/2 V_F
        self.magnetic_helpers(0.5)
        b_cross_a = np.cross(self.magnetic_field_uvector, ptcls.acc)
        b_cross_b_cross_a = np.cross(self.magnetic_field_uvector, b_cross_a)
        ptcls.vel += ptcls.acc * 0.5 * self.dt - self.ccodt/self.omega_c * b_cross_a \
                     + 0.5 * self.dt * self.ssodt * b_cross_b_cross_a

        # epsilon V_B
        self.magnetic_helpers(1.0)
        b_cross_v = np.cross(self.magnetic_field_uvector, ptcls.vel)
        b_cross_b_cross_v = np.cross(self.magnetic_field_uvector, b_cross_v)
        ptcls.vel += - self.sdt * b_cross_v + self.ccodt * b_cross_b_cross_v

        # # epsilon/2 V_F
        self.magnetic_helpers(0.5)
        b_cross_a = np.cross(self.magnetic_field_uvector, ptcls.acc)
        b_cross_b_cross_a = np.cross(self.magnetic_field_uvector, b_cross_a)
        ptcls.vel += ptcls.acc * 0.5 * self.dt - self.ccodt/self.omega_c * b_cross_a \
                     + 0.5 * self.dt * self.ssodt * b_cross_b_cross_a

        # Full step position update
        ptcls.pos += ptcls.vel * self.dt

        # Periodic boundary condition
        enforce_pbc(ptcls.pos, ptcls.pbc_cntr, self.box_lengths)

        # Compute total potential energy and acceleration for second half step velocity update
        potential_energy = self.update_accelerations(ptcls)

        return potential_energy

    def pretty_print(self, frequency, restart, restart_step):
        """Print integrator attributes in a user friendly way."""

        if self.magnetized and self.electrostatic_equilibration:
            print("Type: {}".format(self.magnetic_integrator.__name__))
        else:
            print("Type: {}".format(self.type))

        wp_dt = frequency * self.dt
        print('Time step = {:.6e} [s]'.format(self.dt))
        print('Total plasma frequency = {:.6e} [Hz]'.format(frequency))
        print('w_p dt = {:.4f} ~ 1/{}'.format(wp_dt, int(1.0/wp_dt) ))
        # if potential_type in ['Yukawa', 'EGS', 'Coulomb', 'Moliere']:
        #     # if simulation.parameters.magnetized:
        #     #     if simulation.parameters.num_species > 1:
        #     #         high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
        #     #         low_wc_dt = simulation.parameters.species_cyclotron_frequencies.min() * simulation.integrator.dt
        #     #         print('Highest w_c dt = {:2.4f}'.format(high_wc_dt))
        #     #         print('Smalles w_c dt = {:2.4f}'.format(low_wc_dt))
        #     #     else:
        #     #         high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
        #     #         print('w_c dt = {:2.4f}'.format(high_wc_dt))
        # elif simulation.potential.type == 'QSP':
        #     print('e plasma frequency = {:.6e} [Hz]'.format(simulation.species[0].plasma_frequency))
        #     print('ion plasma frequency = {:.6e} [Hz]'.format(simulation.species[1].plasma_frequency))
        #     print('w_pe dt = {:2.4f}'.format(simulation.integrator.dt * simulation.species[0].plasma_frequency))
        #     if simulation.parameters.magnetized:
        #         if simulation.parameters.num_species > 1:
        #             high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
        #             low_wc_dt = simulation.parameters.species_cyclotron_frequencies.min() * simulation.integrator.dt
        #             print('Electron w_ce dt = {:2.4f}'.format(high_wc_dt))
        #             print('Ions w_ci dt = {:2.4f}'.format(low_wc_dt))
        #         else:
        #             high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
        #             print('w_c dt = {:2.4f}'.format(high_wc_dt))
        # elif simulation.potential.type == 'LJ':
        #     print('Total equivalent plasma frequency = {:1.6e} [Hz]'.format(
        #         simulation.parameters.total_plasma_frequency))
        #     print('w_p dt = {:2.4f}'.format(wp_dt))
        #     if simulation.parameters.magnetized:
        #         if simulation.parameters.num_species > 1:
        #             high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
        #             low_wc_dt = simulation.parameters.species_cyclotron_frequencies.min() * simulation.integrator.dt
        #             print('Highest w_c dt = {:2.4f}'.format(high_wc_dt))
        #             print('Smalles w_c dt = {:2.4f}'.format(low_wc_dt))
        #         else:
        #             high_wc_dt = simulation.parameters.species_cyclotron_frequencies.max() * simulation.integrator.dt
        #             print('w_c dt = {:2.4f}'.format(high_wc_dt))

        # Print Time steps information
        # Check for restart simulations
        if restart in ['production_restart', 'prod_restart']:
            print("Restart step: {}".format(restart_step))
            print('Total production steps = {} \n'
                  'Total production time = {:.4e} [s] ~ {} w_p T_prod '.format(
                self.production_steps,
                self.production_steps * self.dt,
                int(self.production_steps * wp_dt)))
            print('snapshot interval step = {} \n'
                  'snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap'.format(
                self.prod_dump_step,
                self.prod_dump_step * self.dt,
                self.prod_dump_step * wp_dt))
            print('Total number of snapshots = {} '.format( int(self.production_steps /self.prod_dump_step) ) )

        elif restart in ['equilibration_restart', 'eq_restart']:
            print("Restart step: {}".format(restart_step))
            print('Total equilibration steps = {} \n'
                  'Total equilibration time = {:.4e} [s] ~ {} w_p T_eq'.format(
                self.equilibration_steps,
                self.equilibration_steps * self.dt,
                int(self.eq_dump_step * wp_dt)))
            print('snapshot interval step = {} \n'
                  'snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap'.format(
                self.eq_dump_step,
                self.eq_dump_step * self.dt,
                self.eq_dump_step * wp_dt))
            print('Total number of snapshots = {} '.format(int(self.equilibration_steps / self.eq_dump_step)))

        elif restart in ['magnetization_restart', 'mag_restart']:
            print("Restart step: {}".format(restart_step))
            print('Total magnetization steps = {} \n'
                  'Total magnetization time = {:.4e} [s] ~ {} w_p T_mag'.format(
                self.magnetization_steps,
                self.magnetization_steps * self.dt,
                int(self.mag_dump_step * wp_dt)))
            print('snapshot interval step = {} \n'
                  'snapshot interval time = {:.4e} [s] ~ {:.4f} w_p T_snap'.format(
                self.mag_dump_step,
                self.mag_dump_step * self.dt,
                self.mag_dump_step * wp_dt))
            print('Total number of snapshots = {} '.format(int(self.magnetization_steps / self.mag_dump_step)))

        else:
            # Equilibration
            print('\nEquilibration: \nNo. of equilibration steps = {} \n'
                  'Total equilibration time = {:.4e} [s] ~ {} w_p T_eq '.format(
                self.equilibration_steps,
                self.equilibration_steps * self.dt,
                int(self.equilibration_steps * wp_dt)))
            print('snapshot interval step = {} \n'
                  'snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap'.format(
                self.eq_dump_step,
                self.eq_dump_step * self.dt,
                self.eq_dump_step * wp_dt))
            print('Total number of snapshots = {} '.format(int(self.equilibration_steps / self.eq_dump_step)))
            
            # Magnetization
            if self.electrostatic_equilibration:
                print('Electrostatic Equilibration Type: {}'.format(self.type))

                print('\nMagnetization: \nNo. of magnetization steps = {} \n'
                      'Total magnetization time = {:.4e} [s] ~ {} w_p T_mag '.format(
                    self.magnetization_steps,
                    self.magnetization_steps * self.dt,
                    int(self.magnetization_steps * wp_dt)))

                print('snapshot interval step = {} \n'
                      'snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap'.format(
                    self.mag_dump_step,
                    self.mag_dump_step * self.dt,
                    self.mag_dump_step * wp_dt))
                print('Total number of snapshots = {} '.format(int(self.magnetization_steps / self.mag_dump_step)))
            # Production
            print('\nProduction: \nNo. of production steps = {} \n'
                  'Total production time = {:.4e} [s] ~ {} w_p T_prod '.format(
                self.production_steps,
                self.production_steps * self.dt,
                int(self.production_steps * wp_dt)))
            print('snapshot interval step = {} \n'
                  'snapshot interval time = {:.4e} [s] = {:.4f} w_p T_snap'.format(
                self.prod_dump_step,
                self.prod_dump_step * self.dt,
                self.prod_dump_step * wp_dt))
            print('Total number of snapshots = {} '.format(int(self.production_steps / self.prod_dump_step)))


@njit
def enforce_pbc(pos, cntr, BoxVector):
    """ 
    Enforce Periodic Boundary conditions. 

    Parameters
    ----------
    pos: numpy.ndarray
        Particles' positions.

    cntr: numpy.ndarray
        Counter for the number of times each particle get folded back into the main simulation box

    BoxVector: numpy.ndarray
        Box Dimensions.

    """

    # Loop over all particles
    for p in np.arange(pos.shape[0]):
        for d in np.arange(pos.shape[1]):

            # If particle is outside of box in positive direction, wrap to negative side
            if pos[p, d] > BoxVector[d]:
                pos[p, d] -= BoxVector[d]
                cntr[p, d] += 1
            # If particle is outside of box in negative direction, wrap to positive side
            if pos[p, d] < 0.0:
                pos[p, d] += BoxVector[d]
                cntr[p, d] -= 1


@njit
def remove_drift(vel, nums, masses):
    """
    Enforce conservation of total linear momentum. Updates ``particles.vel``

    Parameters
    ----------
    vel: numpy.ndarray
        Particles' velocities.

    nums: numpy.ndarray
        Number of particles of each species.

    masses: numpy.ndarray
        Mass of each species.

    """

    P = np.zeros((len(nums), vel.shape[1]))
    species_start = 0
    for ic in range(len(nums)):
        species_end = species_start + nums[ic]
        P[ic, :] = np.sum(vel[species_start:species_end, :], axis=0) * masses[ic]
        species_start = species_end

    if np.sum(P[:, 0]) > 1e-40 or np.sum(P[:, 1]) > 1e-40 or np.sum(P[:, 2]) > 1e-40:
        # Remove tot momentum
        species_start = 0
        for ic in range(len(nums)):
            species_end = species_start + nums[ic]
            vel[species_start:species_end, :] -= P[ic, :] / (float(nums[ic]) * masses[ic])
            species_start = species_end



"""
Module for handling the Particle-Mesh part of the force and potential calculation.
"""

import numpy as np
from numba import jit, njit
import pyfftw

# These "ignore" are needed because numba does not support pyfftw yet
from numba.core.errors import NumbaWarning, NumbaDeprecationWarning, NumbaPendingDeprecationWarning
import warnings
warnings.simplefilter('ignore', category=NumbaDeprecationWarning)
warnings.simplefilter('ignore', category=NumbaWarning)
warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)


@njit
def force_optimized_green_function(box_lengths, mesh_sizes, aliases, p, constants):
    """
    Calculate the Optimized Green Function given by eq.(22) of Ref. [Stern2008].

    Parameters
    ----------

    mesh_sizes : numpy.ndarray
        number of mesh points in x,y,z

    aliases : numpy.ndarray
        number of aliases in each direction

    box_lengths : numpy.ndarray
        Length of simulation's box in each direction

    p : int
        Charge assignment order (CAO)

    constants : numpy.ndarray
        Screening parameter, Ewald parameter, 4 pi eps0.

    Returns
    -------
    G_k : numpy.ndarray
        optimal Green Function

    kx_v : numpy.ndarray
       array of reciprocal space vectors along the x-axis

    ky_v : numpy.ndarray
       array of reciprocal space vectors along the y-axis

    kz_v : numpy.ndarray
       array of reciprocal space vectors along the z-axis

    PM_err : float
        Error in the force calculation due to the optimized Green's function. eq.(28) of Ref. [Dharuman2017]

    PP_err : float
        Error in the force calculation due to the distance cutoff. eq.(30) of Ref. [Stern2008]

    """
    kappa = constants[0]
    Gew = constants[1]
    fourpie0 = constants[2]

    h_array = box_lengths / mesh_sizes

    kappa_sq = kappa * kappa
    Gew_sq = Gew * Gew

    G_k = np.zeros((mesh_sizes[2], mesh_sizes[1], mesh_sizes[0]))

    nz_mid = mesh_sizes[2]/2 if np.mod(mesh_sizes[2], 2) == 0 else (mesh_sizes[2] - 1)/2
    ny_mid = mesh_sizes[1]/2 if np.mod(mesh_sizes[1], 2) == 0 else (mesh_sizes[1] - 1)/2
    nx_mid = mesh_sizes[0]/2 if np.mod(mesh_sizes[0], 2) == 0 else (mesh_sizes[0] - 1)/2

    # nx_v = np.arange(mesh_sizes[0]).reshape((1, mesh_sizes[0]))
    # ny_v = np.arange(mesh_sizes[1]).reshape((mesh_sizes[1], 1))
    # nz_v = np.arange(mesh_sizes[2]).reshape((mesh_sizes[2], 1, 1))
    # Dev Note:
    # The above three lines where giving a problem with Numba in Windows only.
    # I replaced them with the ones below. I don't know why it was giving a problem.
    nx_v = np.zeros((1,mesh_sizes[0]), dtype=np.int64)
    nx_v[0,:] = np.arange(mesh_sizes[0])

    ny_v = np.zeros((mesh_sizes[1], 1), dtype=np.int64)
    ny_v[:,0] = np.arange(mesh_sizes[1])
    
    nz_v = np.zeros((mesh_sizes[2], 1, 1), dtype=np.int64)
    nz_v[:,0, 0] = np.arange(mesh_sizes[2])
    
    kx_v = 2.0 * np.pi * (nx_v - nx_mid) / box_lengths[0]
    ky_v = 2.0 * np.pi * (ny_v - ny_mid) / box_lengths[1]
    kz_v = 2.0 * np.pi * (nz_v - nz_mid) / box_lengths[2]

    PM_err = 0.0

    four_pi = 4.0 * np.pi if fourpie0 == 1.0 else 4.0 * np.pi / fourpie0
    two_pi = 2.0 * np.pi

    for nz in range(mesh_sizes[2]):
        nz_sh = nz - nz_mid
        kz = two_pi * nz_sh / box_lengths[2]

        for ny in range(mesh_sizes[1]):
            ny_sh = ny - ny_mid
            ky = two_pi * ny_sh / box_lengths[1]

            for nx in range(mesh_sizes[0]):
                nx_sh = nx - nx_mid
                kx = two_pi * nx_sh / box_lengths[0]

                k_sq = kx * kx + ky * ky + kz * kz

                if k_sq != 0.0:

                    U_k_sq = 0.0
                    U_G_k = 0.0

                    # Sum over the aliases
                    for mz in range(-aliases[2], aliases[2] + 1):
                        kz_M = two_pi * (nz_sh + mz * mesh_sizes[2]) / box_lengths[2]
                        U_kz_M = np.sin(0.5 * kz_M * h_array[2]) / (0.5 * kz_M * h_array[2]) if kz_M != 0.0 else 1.0

                        for my in range(-aliases[1], aliases[1] + 1):
                            ky_M = two_pi * (ny_sh + my * mesh_sizes[1]) / box_lengths[1]
                            U_ky_M = np.sin(0.5 * ky_M * h_array[1]) / (0.5 * ky_M * h_array[1]) if ky_M != 0.0 else 1.0

                            for mx in range(-aliases[0], aliases[0] + 1):
                                kx_M = two_pi * (nx_sh + mx * mesh_sizes[0]) / box_lengths[0]
                                U_kx_M = np.sin(0.5 * kx_M * h_array[0]) / (0.5 * kx_M * h_array[0]) if kx_M != 0.0 else 1.0

                                k_M_sq = kx_M * kx_M + ky_M * ky_M + kz_M * kz_M

                                U_k_M = (U_kx_M * U_ky_M * U_kz_M) ** p
                                U_k_M_sq = U_k_M * U_k_M

                                G_k_M = four_pi * np.exp(-0.25 * (kappa_sq + k_M_sq) / Gew_sq) / (kappa_sq + k_M_sq)

                                k_dot_k_M = kx * kx_M + ky * ky_M + kz * kz_M

                                U_G_k += (U_k_M_sq * G_k_M * k_dot_k_M)
                                U_k_sq += U_k_M_sq

                    # eq.(22) of Ref.[Dharuman2017]_
                    G_k[nz, ny, nx] = U_G_k / ((U_k_sq ** 2) * k_sq)
                    Gk_hat = four_pi * np.exp(-0.25 * (kappa_sq + k_sq) / Gew_sq) / (kappa_sq + k_sq)

                    # eq.(28) of Ref.[Dharuman2017]_
                    PM_err += Gk_hat * Gk_hat * k_sq - U_G_k ** 2 / ((U_k_sq ** 2) * k_sq)

    PM_err = np.sqrt(PM_err) / np.prod(box_lengths) ** (1. / 3.)

    return G_k, kx_v, ky_v, kz_v, PM_err


@njit
def assgnmnt_func(cao, x):
    """ 
    Calculate the charge assignment function as given in Ref. [Deserno1998].
    
    Parameters
    ----------
    cao : int
        Charge assignment order.

    x : float
        Distance to closest mesh point if cao is even.
    
    Returns
    ------
    W : numpy.ndarray
        Charge Assignment Function. 

    """
    W = np.zeros(cao)

    if cao == 1:

        W[0] = 1

    elif cao == 2:

        W[0] = 0.5 * (1. - 2. * x)
        W[1] = 0.5 * (1. + 2. * x)

    elif cao == 3:

        W[0] = (1. - 4. * x + 4. * x ** 2) / 8.
        W[1] = (3. - 4. * x ** 2) / 4.
        W[2] = (1. + 4. * x + 4. * x ** 2) / 8.

    elif cao == 4:

        W[0] = (1. - 6. * x + 12. * x ** 2 - 8. * x ** 3) / 48.
        W[1] = (23. - 30. * x - 12. * x ** 2 + 24. * x ** 3) / 48.
        W[2] = (23. + 30. * x - 12. * x ** 2 - 24. * x ** 3) / 48.
        W[3] = (1. + 6. * x + 12. * x ** 2 + 8. * x ** 3) / 48.

    elif cao == 5:

        W[0] = (1. - 8. * x + 24. * x ** 2 - 32. * x ** 3 + 16. * x ** 4) / 384.
        W[1] = (19. - 44. * x + 24. * x ** 2 + 16. * x ** 3 - 16. * x ** 4) / 96.
        W[2] = (115. - 120. * x ** 2 + 48. * x ** 4) / 192.
        W[3] = (19. + 44. * x + 24. * x ** 2 - 16. * x ** 3 - 16. * x ** 4) / 96.
        W[4] = (1. + 8. * x + 24. * x ** 2 + 32. * x ** 3 + 16. * x ** 4) / 384.

    elif cao == 6:
        W[0] = (1. - 10. * x + 40. * x ** 2 - 80. * x ** 3 + 80. * x ** 4 - 32. * x ** 5) / 3840.
        W[1] = (237. - 750. * x + 840. * x ** 2 - 240. * x ** 3 - 240. * x ** 4 + 160. * x ** 5) / 3840.
        W[2] = (841. - 770. * x - 440. * x ** 2 + 560. * x ** 3 + 80. * x ** 4 - 160. * x ** 5) / 1920.
        W[3] = (841. + 770. * x - 440. * x ** 2 - 560. * x ** 3 + 80. * x ** 4 + 160. * x ** 5) / 1920.
        W[4] = (237. + 750. * x + 840. * x ** 2 + 240. * x ** 3 - 240. * x ** 4 - 160. * x ** 5) / 3840.
        W[5] = (1. + 10. * x + 40. * x ** 2 + 80. * x ** 3 + 80. * x ** 4 + 32. * x ** 5) / 3840.

    elif cao == 7:

        W[0] = (1. - 12. * x + 60. * x * 2 - 160. * x ** 3 + 240. * x ** 4 - 192. * x ** 5 + 64. * x ** 6) / 46080.

        W[1] = (361. - 1416. * x + 2220. * x ** 2 - 1600. * x ** 3 + 240. * x ** 4
                + 384. * x ** 5 - 192. * x ** 6) / 23040.

        W[2] = (10543. - 17340. * x + 4740. * x ** 2 + 6880. * x ** 3 - 4080. * x ** 4
                - 960. * x ** 5 + 960. * x ** 6) / 46080.

        W[3] = (5887. - 4620. * x ** 2 + 1680. * x ** 4 - 320. * x ** 6) / 11520.

        W[4] = (10543. + 17340. * x + 4740. * x ** 2 - 6880. * x ** 3 - 4080. * x ** 4
                + 960. * x ** 5 + 960. * x ** 6) / 46080.

        W[5] = (361. + 1416. * x + 2220. * x ** 2 + 1600. * x ** 3 + 240. * x ** 4
                - 384. * x ** 5 - 192. * x ** 6) / 23040.

        W[6] = (1. + 12. * x + 60. * x ** 2 + 160. * x ** 3 + 240. * x ** 4 + 192. * x ** 5 + 64. * x ** 6) / 46080.

    return W


@njit
def calc_charge_dens(pos, charges, N, cao, mesh_sz, h_array):
    """ 
    Assigns Charges to Mesh Points.

    Parameters
    ----------
    h_array: numpy.ndarray
        Distances between mesh points per dimension.

    mesh_sz: numpy.ndarray
        Mesh points per direction.

    pos: numpy.ndarray
        Particles' positions.
    
    charges: numpy.ndarray
        Particles' charges.
    
    N: int
        Number of particles.

    cao: int
        Charge assignment order.

    Returns
    -------
    rho_r: numpy.ndarray
        Charge density distributed on mesh.

    """

    rho_r = np.zeros((mesh_sz[2], mesh_sz[1], mesh_sz[0]))

    # Mid point calculation
    if cao % 2 == 0:
        # Choose the midpoint between the two closest mesh point to the particle's position
        mid = 0.5
        pshift = int(cao / 2 - 1)
    else:
        # Choose the mesh point closes to the particle
        mid = 0.0
        pshift = int(cao / float(2.0))

    for ipart in range(N):

        # ix = x-coord of the (left) closest mesh point
        # (ix + 0.5)*h_array[0] = midpoint between the two mesh points closest to the particle
        # x = the difference between the particle's position and the midpoint
        # Rescale

        ix = int(pos[ipart, 0] / h_array[0])
        x = pos[ipart, 0] / h_array[0] - (ix + mid)

        iy = int(pos[ipart, 1] / h_array[1])
        y = pos[ipart, 1] / h_array[1] - (iy + mid)

        iz = int(pos[ipart, 2] / h_array[2])
        z = pos[ipart, 2] / h_array[2] - (iz + mid)

        wx = assgnmnt_func(cao, x)
        wy = assgnmnt_func(cao, y)
        wz = assgnmnt_func(cao, z)

        izn = iz - pshift  # min. index along z-axis

        for g in range(cao):

            # if izn < 0:
            r_g = izn + mesh_sz[2] * (izn < 0) - mesh_sz[2] * (izn > (mesh_sz[2] - 1))
            # elif izn > (mesh_sz[2] - 1):
            #     r_g = izn - mesh_sz[2]
            # else:
            #     r_g = izn

            iyn = iy - pshift  # min. index along y-axis

            for i in range(cao):

                r_i = iyn + mesh_sz[1] * (iyn < 0) - mesh_sz[1] * (iyn > (mesh_sz[1] - 1))

                # if iyn < 0:
                #     r_i = iyn + mesh_sz[1]
                # elif iyn > (mesh_sz[1] - 1):
                #     r_i = iyn - mesh_sz[1]
                # else:
                #     r_i = iyn

                ixn = ix - pshift  # min. index along x-axis

                for j in range(cao):

                    r_j = ixn + mesh_sz[0] * (ixn < 0) - mesh_sz[0] * (ixn > (mesh_sz[0] - 1))

                    # if ixn < 0:
                    #     r_j = ixn + mesh_sz[0]
                    # elif ixn > (mesh_sz[0] - 1):
                    #     r_j = ixn - mesh_sz[0]
                    # else:
                    #     r_j = ixn

                    rho_r[r_g, r_i, r_j] += charges[ipart] * wz[g] * wy[i] * wx[j]

                    ixn += 1

                iyn += 1

            izn += 1

    return rho_r


@njit
def calc_field(phi_k, kx_v, ky_v, kz_v):
    """ 
    Calculates the Electric field in Fourier space.

    Parameters
    ----------
    phi_k : numpy.ndarray
        3D array of the Potential.

    kx_v : numpy.ndarray
        3D array containing the values of kx.

    ky_v : numpy.ndarray
        3D array containing the values of ky.

    kz_v : numpy.ndarray
        3D array containing the values of kz.
    
    Returns
    -------
    E_kx : numpy.ndarray
       Electric Field along kx-axis.

    E_ky : numpy.ndarray
       Electric Field along ky-axis.

    E_kz : numpy.ndarray
       Electric Field along kz-axis.
    
    """

    E_kx = -1j * kx_v * phi_k
    E_ky = -1j * ky_v * phi_k
    E_kz = -1j * kz_v * phi_k

    return E_kx, E_ky, E_kz


@njit
def calc_acc_pm(E_x_r, E_y_r, E_z_r, pos, charges, N, cao, masses, mesh_sz, h_array):
    """ 
    Calculates the long range part of particles' accelerations. 
    
    Parameters
    ----------
    E_x_r : numpy.ndarray
        Electric field along x-axis.

    E_y_r : numpy.ndarray
        Electric field along y-axis.
    
    E_z_r : numpy.ndarray
        Electric field along z-axis.
    
    pos : numpy.ndarray
        Particles' positions.
    
    charges : numpy.ndarray
        Particles' charges.
    
    N : int
        Number of particles.

    cao : int
        Charge assignment order.
    
    masses : numpy.ndarray
        Particles' masses.


    Returns
    -------

    acc : numpy.ndarray
          Acceleration from Electric Field.

    """
    E_x_p = np.zeros(N)
    E_y_p = np.zeros(N)
    E_z_p = np.zeros(N)

    acc = np.zeros_like(pos)

    # Mid point calculation
    if cao % 2 == 0:
        # Choose the midpoint between the two closest mesh point to the particle's position
        mid = 0.5
        # Number of points to the left of the chosen one
        pshift = int(cao / 2 - 1)
    else:
        # Choose the mesh point closes to the particle
        mid = 0.0
        # Number of points to the left of the chosen one
        pshift = int(cao / float(2.0))

    for ipart in range(N):

        ix = int(pos[ipart, 0] / h_array[0])
        x = pos[ipart, 0] / h_array[0] - (ix + mid)

        iy = int(pos[ipart, 1] / h_array[1])
        y = pos[ipart, 1] / h_array[1] - (iy + mid)

        iz = int(pos[ipart, 2] / h_array[2])
        z = pos[ipart, 2] / h_array[2] - (iz + mid)

        wx = assgnmnt_func(cao, x)
        wy = assgnmnt_func(cao, y)
        wz = assgnmnt_func(cao, z)

        izn = iz - pshift  # min. index along z-axis

        for g in range(cao):
            #
            # if izn < 0:
            #     r_g = izn + mesh_sz[2]
            # elif izn > (mesh_sz[2] - 1):
            #     r_g = izn - mesh_sz[2]
            # else:
            #     r_g = izn

            r_g = izn + mesh_sz[2] * (izn < 0) - mesh_sz[2] * (izn > (mesh_sz[2] - 1))

            iyn = iy - pshift  # min. index along y-axis

            for i in range(cao):

                # if iyn < 0:
                #     r_i = iyn + mesh_sz[1]
                # elif iyn > (mesh_sz[1] - 1):
                #     r_i = iyn - mesh_sz[1]
                # else:
                #     r_i = iyn
                r_i = iyn + mesh_sz[1] * (iyn < 0) - mesh_sz[1] * (iyn > (mesh_sz[1] - 1))

                ixn = ix - pshift  # min. index along x-axis

                for j in range(cao):

                    r_j = ixn + mesh_sz[0] * (ixn < 0) - mesh_sz[0] * (ixn > (mesh_sz[0] - 1))

                    # if ixn < 0:
                    #     r_j = ixn + mesh_sz[0]
                    # elif ixn > (mesh_sz[0] - 1):
                    #     r_j = ixn - mesh_sz[0]
                    # else:
                    #     r_j = ixn

                    q_over_m = charges[ipart] / masses[ipart]
                    E_x_p[ipart] += q_over_m * E_x_r[r_g, r_i, r_j] * wz[g] * wy[i] * wx[j]
                    E_y_p[ipart] += q_over_m * E_y_r[r_g, r_i, r_j] * wz[g] * wy[i] * wx[j]
                    E_z_p[ipart] += q_over_m * E_z_r[r_g, r_i, r_j] * wz[g] * wy[i] * wx[j]

                    ixn += 1

                iyn += 1

            izn += 1

    acc[:, 0] = E_x_p
    acc[:, 1] = E_y_p
    acc[:, 2] = E_z_p

    return acc


# FFTW version
@jit  # Numba does not support pyfftw yet, however, this decorator still speeds up the function.
def update(pos, charges, masses, mesh_sizes, box_lengths, G_k, kx_v, ky_v, kz_v, cao):
    """ 
    Calculate the long range part of particles' accelerations.

    Parameters
    ----------
    box_lengths: numpy.ndarray
        Box length in each direction.

    mesh_sizes: numpy.ndarray
        Mesh points per direction.

    pos: numpy.ndarray
        Particles' positions.

    charges: numpy.ndarray
        Particles' charges.

    masses: numpy.ndarray
        Particles' masses.

    G_k : numpy.ndarray
        Optimized Green's function.

    kx_v : numpy.ndarray
        Array of kx values.

    ky_v : numpy.ndarray
        Array of ky values.

    kz_v : numpy.ndarray
        Array of kz values.
    
    cao : int
        Charge order parameter.

    Returns
    -------
    U_f : float
        Long range part of the potential.

    acc_f : numpy.ndarray
        Long range part of particles' accelerations.

    """
    # number of particles
    N = pos.shape[0]
    # Mesh spacings = h_x, h_y, h_z
    mesh_spacings = box_lengths/mesh_sizes
    # Calculate charge density on mesh
    rho_r = calc_charge_dens(pos, charges, N, cao, mesh_sizes, mesh_spacings)
    # Prepare for fft
    fftw_n = pyfftw.builders.fftn(rho_r)
    # Calculate fft
    rho_k_fft = fftw_n()

    # Shift the DC value at the center of the ndarray
    rho_k = np.fft.fftshift(rho_k_fft)

    # Potential from Poisson eq.
    phi_k = G_k * rho_k

    # Charge density
    rho_k_real = np.real(rho_k)
    rho_k_imag = np.imag(rho_k)
    rho_k_sq = rho_k_real * rho_k_real + rho_k_imag * rho_k_imag

    # Long range part of the potential
    U_f = 0.5 * np.sum(rho_k_sq * G_k) / np.prod(box_lengths)

    # Calculate the Electric field's component on the mesh
    E_kx, E_ky, E_kz = calc_field(phi_k, kx_v, ky_v, kz_v)

    # Prepare for fft. Shift the DC value back to its original position that is [0, 0, 0]
    E_kx_unsh = np.fft.ifftshift(E_kx)
    E_ky_unsh = np.fft.ifftshift(E_ky)
    E_kz_unsh = np.fft.ifftshift(E_kz)

    # Prepare and compute IFFT
    ifftw_n = pyfftw.builders.ifftn(E_kx_unsh)
    E_x = ifftw_n()
    ifftw_n = pyfftw.builders.ifftn(E_ky_unsh)
    E_y = ifftw_n()
    ifftw_n = pyfftw.builders.ifftn(E_kz_unsh)
    E_z = ifftw_n()

    # I am worried that this normalization is not needed
    E_x_r = np.real(E_x) / np.prod(mesh_spacings)
    E_y_r = np.real(E_y) / np.prod(mesh_spacings)
    E_z_r = np.real(E_z) / np.prod(mesh_spacings)

    acc_f = calc_acc_pm(E_x_r, E_y_r, E_z_r, pos, charges, N, cao, masses, mesh_sizes, mesh_spacings)

    return U_f, acc_f
